import logging
import json
import logging
import pandas as pd
from datetime import datetime
import pathlib
from collections import namedtuple
import os
from typing_extensions import final
from unit_converter.converter import convert, converts
from enum import Enum
from assertpy import assert_that

NoneType = type(None)

my_logger = logging.getLogger(__name__)

class execution_information:
    """
    This class maintains the test details json structure as:
    {
    "Execution_ID": "", - autogenerate at the moment the object is initialized
    "DUT_identifier": "", - uniqe identifier of the DUT
    "Chip_version": "", - identifies the version of the chip
    "Operator": "", - operator logged in the computer
    "Test_bench_ID": "", - identifies the test setup
    "Execution_temperature": "", - temperature of the chip
    "SVN_revision": "", - revision id of the test script from the repository
    "Execution_started": "" - timestamp when the execution is started
    "Additional_information": "" - optional user given information about the execution
    }
    This information will be included in the log and result file
    
    Functionality:
    - gets information from .json file
    - write information into log file and result file
    - shares information to other modules
    """
    def __init__(self):
        """
        initializes the object
        """
        self.__details_dict = {}


    def get_information_from_file(self):
        """
        This file reads the .json file with the test details and output them into the console window as well as the log file
        In addition it stores json data in a dictionary
        """
        with open(f'{os.getcwd()}\\resources\\curr_test_details.json', 'r') as test_details_file:
            self.__details_dict = json.load(test_details_file)
        return self.__details_dict


    def get_test_info_by_name(self, info_name):
        """
        this function returns one specific data from the dictionary that stores the details of the test
        """
        return self.__details_dict[info_name]


class failure_collector:
    """
    This class collects the failures occured during exection into an error log file
    Failure log file is generated into the results folder named as failures_[chip_ID]_[execution_ID].log
    """
    def __init__(self):
        """
        This method initializes the object and creates the log file
        The execution ID is obtained from curr_test_details.json from the resources folder
        """
        with open(f'{os.getcwd()}\\resources\\curr_test_details.json', 'r') as test_details_file:
            execution_info = json.load(test_details_file)
            execution_ID = execution_info['Execution_ID']
            chip_ID = execution_info['DUT_identifier']
        self._details_file_path = f'{os.getcwd()}\\results\\failures_{chip_ID}_{execution_ID}.log'


    def add_failure(self, test_ID, value, limit_L, limit_H, unit, param_name, temperature, filename: str=None):
        """
        This method appends a new line to the failure log file with timestamp
        It records the details of the failure:
        - test_ID - to identify the test step where the failure occured
        - value - the value that made the test fail
        - limit_L as lower limit
        - limit_H as upper limit
        - unit to indicate the dimension of the measured value
        - param_name that describes the parameter that is represented by the value
        - temperature of the measurement
        """
        timestamp = datetime.strftime(datetime.now(), '%Y-%m-%d_%H-%M-%S')
        if pathlib.Path(self._details_file_path).exists():
            mode = 'a'
        else:
            mode = 'w'
        with open(self._details_file_path, mode) as my_failure_log:
            if filename != None:
                failure_text = f'\n{timestamp} #{test_ID}: {value} {unit} ({limit_L} / {limit_H}) - {param_name} @{temperature}deg-C -> {filename}'
            else:
                failure_text = f'\n{timestamp} #{test_ID}: {value} {unit} ({limit_L} / {limit_H}) - {param_name} @{temperature}deg-C'
            my_failure_log.write(
                failure_text
            )


class details_grabber:
    """
    This class is responsible of collecting the information in case of iteration over a higher number of tests
    It opens a .json file from the resources folder
    """
    @classmethod
    def get_test_group_details(self, filename, section=None):
        """
        This method opens a .json file defined by the 'filename' and imports the section that is defined by the 'section' parameter
        In case the sectiion parameter is not given the function returns the whole file content as a list of dictionaries
        """
        with open(f'{os.getcwd()}\\resources\\{filename}', 'r') as details_file:
            file_content = json.load(details_file)
        if section:
            return file_content[section]
        else:
            return file_content


    @classmethod
    def get_test_group_detail(self, filename: str, section: str, detail_name: str='test_ID'):
        """
        This method returns a list of details json file's section and builds a list of the field that is specified by the detail name

        Args:
            filename (str): details file path
            section (str): section of the json file
            detail_name (str): name of the field to be collected like test_ID

        Returns:
            [type]: [description]
        """
        with open(f'{os.getcwd()}\\resources\\{filename}', 'r') as details_file:
            file_content = json.load(details_file)
        
        my_section = file_content[section]

        ids = []
        for item in my_section:
            ids.append(item[detail_name])
        return ids


class TrimFileHandler:
    """
    This class is responsible of trim file and value management
    """
    def __init__(self, parameter_name: str, unit: str) -> object:
        """__init__ 
        This method initializes a new trim dictionary

        Args:
            parameter_name (str): name of the parameer to be trimmed
            unit (str): unit of the physical value to be trimmed

        Returns:
            object: new trim object
        """
        self.trim_dict = {
            "parameter_name": parameter_name,
            "unit": unit,
            "flash_trim": 0,
            "quasy_trim": 0,
            "table": []
        }


    def addTrimValue(self, register_value: int, physical_value: float):
        """add_trim_value 
        This method adds a new element to a trim table if the table

        Args:
            parameter_name (str): name of the trim parameter
            register_value (int): register value set
            physical_value (float): physical value measured
            unit (str): unit of the physical value
        """
        self.trim_dict['table'].append([register_value, physical_value])


    def setFlashTrim(self, flash_trim_value: int):
        """set_flash_trim 
        Records the trim value that was already in the chip

        Args:
            flash_trim_value (int): the trim value read out from the chip
        """
        self.trim_dict['flash_trim'] = flash_trim_value


    def getFlashTrim(self, trim_name: str=None) -> int:
        """getFlashTrim 
        Returns the flash trim value. If trim name is not defined then it will return the flash trim of the actual object

        Args:
            trim_name (str, optional): name of the parameter. Defaults to None.

        Returns:
            int: register value from file
        """
        if trim_name !=None:
            with open(f'{os.getcwd()}\\resources\\{trim_name}_trim.json', 'r') as trimfile:
                self.trim_dict = json.load(trimfile)
        return self.trim_dict['flash_trim']


    def saveTrimFile(self):
        """saveTrimFile 
        This method saves the trim data recorded into the object into a file
        """
        trim_name = self.trim_dict['parameter_name']
        trim_file_path = f'{os.getcwd()}\\resources\\{trim_name}_trim.json'
        with open(trim_file_path, 'w') as trim_file:
            json.dump(self.trim_dict, trim_file, indent=5)


    def defineQuasyTrim(self, quasy_trim_value: int):
        """defineQuasyTrim 
        This method sets the quasy trim value for the parameter

        Args:
            quasy_trim_value (int): register value belonging to adjusted physical value
        """
        self.trim_dict['quasy_trim'] = quasy_trim_value


    def getQuasyTrimValue(self, trim_name: str=None) -> int:
        """getQuasyTrimValue 
        Returns the quasy trim value. If trim name is not defined then it will return the quasy trim of the actual object

        Args:
            trim_name (str, optional): name of the parameter. Defaults to None.

        Returns:
            int: quasy trim value from file
        """
        if trim_name !=None:
            with open(f'{os.getcwd()}\\resources\\{trim_name}_trim.json', 'r') as trimfile:
                self.trim_dict = json.load(trimfile)
        return self.trim_dict['quasy_trim']

    
    def getTrimValue(self, nominal: float, accuracy: float, unit: str, trim_name: str=None) -> int:
        """get_trim_value 
        This is the function to be called from the sequence.
        It returns the register value that sets the physical value closes to the value defined by nominal

        Args:
            nominal (float): desired physical value
            accuracy (float): tolerance of the desired physical value
            unit (str): unit of the nominal and accuracy
            trim_name (str): name of the parameter to be trimmed

        Returns:
            int: the register value to be used
        """
        if trim_name !=None:
            with open(f'{os.getcwd()}\\resources\\{trim_name}_trim.json', 'r') as trimfile:
                self.trim_dict = json.load(trimfile)
        self.parameter_name = self.trim_dict['parameter_name']
        self.unit = self.trim_dict['unit']
        self.flash_trim = self.trim_dict['flash_trim']
        self.quasy_trim = self.trim_dict['quasy_trim']
        self.table = self.trim_dict['table']
        return self.__getClosestValue(
            nominal=nominal,
            accuracy=accuracy,
            input_unit=unit,
            table=self.table,
            table_unit=self.unit
        )


    def getCompleteTrimFileInfo(self, trim_name: str=None) -> str:
        if trim_name !=None:
            with open(f'{os.getcwd()}\\resources\\{trim_name}_trim.json', 'r') as trimfile:
                self.trim_dict = json.load(trimfile)
        trim_string = ''
        for key in self.trim_dict.keys():
            if key != 'table':
                act_string = f'{key}: {self.trim_dict[key]}'
                my_logger.info(act_string)
                trim_string = trim_string + act_string + '\n'
        return trim_string


    @classmethod
    def __getClosestValue(self, nominal: float, accuracy: float, input_unit: str, table: list, table_unit: str) -> int:
        """__getClosestValue 
        This method selects the value of the table that fullfills the requirements defined by the nominal and accuracy

        Args:
            nominal (float): desired physical value
            accuracy (float): tolerance of the desired physical value
            input_unit (str): unit of nominval value and accuracy
            table (list): trim calibration table that is a list of float value pairs (list)
            table_unit (str): unit that is applied in the calibraiton table

        Returns:
            int: the register value that fulfills the requirements of nominal and accuracy, None is returned if no value can be found
        """
        my_logger.debug(input_unit)
        my_logger.debug(table_unit)
        conv_nominal = float(convert(f'{nominal} {input_unit}', f'{table_unit}'))
        conv_accuracy = float(convert(f'{accuracy} {input_unit}', f'{table_unit}'))
        my_logger.debug(f'converted value: {conv_nominal}')
        possible_values = [[item[0], abs(item[1] - conv_nominal)] for item in table if abs(item[1] - conv_nominal) < conv_accuracy]
        if len(possible_values) > 0:
            my_logger.debug(possible_values)
            temp_closest = possible_values[0]
            for item in possible_values:
                temp_closest = item if item[1] < temp_closest[1] else temp_closest
            result = temp_closest[0]
        else:
            result = None
        return result


class result_file_writer:
    """
    This class maintains the result file generated during the execution
    The result file is a spreadsheet file the organizes the content of the log files into a table for better overview

    The result file maintains columns:
        - test_ID - the unique identifier of the test
        - parameter_name - describes the parameter
        - min - lower limit of the measurement
        - max - upper limit ot the measurement
        - typ - the typical value that is expected
        - unit - the unit of the limits and the measured value
        - measured_value - the value that is obtained by the test execution
        - result - can be 'PASS' or 'FAIL'

    
    The result file has the same filename as the log file
    Result file is generated in the results folder with the extension of .csv
    """


    def __init__(self, log_file_name):
        """
        This method initializes the object and creates the result file that has the same name as the log file
        It also creates the header for the file
        """
        headers_list = [
            'test_ID',
            'parameter_name',
            'min',
            'max',
            'typ',
            'unit',
            'measured_value',
            'result'
        ]
        self._result_file_path = f'{os.getcwd()}\\results\\{log_file_name}.csv'
        with open(self._result_file_path, 'w') as result_file:
            result_file.write(
                ';'.join(headers_list)
            )


    def append_result(
        self,
        test_ID,
        parameter_name,
        measured_value,
        minimum=0,
        maximum=0,
        typ=0,
        unit=''
    ):
        """
        This function is executed after the evaluation of the assertion is done and appends a row to the result file
        """
        result = self._evaluate(measured_value, minimum, maximum, unit)

        record = [test_ID, parameter_name, str(minimum), str(maximum), str(typ), unit, str(measured_value), result]
        with open(self._result_file_path, 'a') as result_file:
            to_be_written = '\n' + ';'.join(record)
            my_logger.debug(to_be_written)
            result_file.write(
                to_be_written
            )


    @classmethod
    def _evaluate(self, Value, Minimum, Maximum, Unit: str='') -> bool:
        """
        This method evaluates the test and returns true if test is passed based on the given parameters

        Returns:
            bool: this is true if the test is passed
        """
        eval_type = str(type(Value)).replace("<class '", '').replace("'>", '')
        eval_method = getattr(self, f'_eval_{eval_type}')
        if type(Minimum) == type(None) or type(Maximum) == type(None):
            return 'TBD'
        else:
            eval = eval_method(
                Value,
                Minimum,
                Maximum,
                Unit
            )
            ret_eval = 'PASSED' if eval else 'FAILED'
            return ret_eval


    @classmethod
    def _eval_str(self, value: str, minimum: str, maximum: str, unit: str='hex') -> bool:
        """
        This method evaluates numbers flattened into string like hexadecimal string

        Args:
            value (str): string value
            min (str): string lower limit
            max (str): string upper limit
            unit (str): unit defines the format of the number formatted into string as:
                - hex : hexadecimal
                - bin : binary
                - dec : decimal

        Returns:
            bool: returns true if value is between the limits
        """
        my_conversion = getattr(self, f'_convert_{unit}')
        conv_value = my_conversion(value)
        conv_min = my_conversion(minimum)
        conv_max = my_conversion(maximum)
        return (conv_value >= conv_min and conv_value <= conv_max)


    @classmethod
    def _convert_hex(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 16)


    @classmethod
    def _convert_bin(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 2)


    @classmethod
    def _convert_dec(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 10)


    @classmethod
    def _remove_type_marker(self, string_in: str) -> str:
        """
        This method removes the possible type markers from the beginning of the numeric strings like 0x, 0b, 0d
        It also removes extra spaces from the string.

        Args:
            string_in (str): [description]

        Returns:
            str: [description]
        """
        my_string = string_in.replace('0x', '')
        my_string = my_string.replace('0b', '')
        my_string = my_string.replace('0d', '')
        my_string = my_string.replace('0u', '')
        my_string = my_string.replace(' ', '')
        return my_string


    @classmethod
    def _eval_float(self, value: float, minimum: float, maximum: float, unit: str='V') -> bool:
        """
        This method evaluates a result that is floating point numeric

        Args:
            value (float): value
            min (float): lower limit
            max (float): upper limit
            unit (str, optional): unit of the physical quantity. Defaults to 'V'.

        Returns:
            bool: returns true if value is between the limits
        """
        return (value >= minimum and value <= maximum)


    @classmethod
    def _eval_int(self, value: int, minimum: int, maximum: int, unit: str='V') -> bool:
        """
        This method evaluates a result that is integer

        Args:
            value (int): value
            min (int): lower limit
            max (int): upper limit
            unit (str, optional): unit of the physical quantity. Defaults to 'V'.

        Returns:
            bool: returns true if value is between the limits
        """
        return (value >= minimum and value <= maximum)





class ResultHandler:
    """
    This class is responsible of managing the details and results of a specific test
    """

    class ResultEval(Enum):
        """
        This enumeration defines the possible outcome of a test result evaluation
        """
        FAILED = 0
        SKIPPED = 1
        NOT_EVALUATED = 2
        PASSED = 3

    

    results = []
    primary_results = []
    result = ResultEval.NOT_EVALUATED


    @classmethod
    def ClearResults(cls):
        cls.results = []
        cls.primary_results = []
        cls.result = cls.ResultEval.NOT_EVALUATED


    @classmethod
    def addResult(
        cls, 
        measurement_result,
        parameter_name: str=None,
        minimum=None,
        maximum=None,
        unit: str=None,
        typ=None,
        remark: str=None,
        dump: bool=False,
        result_write_object: result_file_writer=None
    ):
        """
        This method adds a result to the test.
        The result added by this mathod will be recorded but it is not managed as required to PASS the test.

        Args:
            measurement_result (str|float|int): measurement result value
            parameter_name (str, optional): name of ther parameter that the measurement result belongs to. Defaults to None.
            min (str|float|int, optional): lower limit. Defaults to None.
            max (str|float|int, optional): upper limit. Defaults to None.
            typ (str|float|int, optional): typical value. Defaults to None.
            unit (str, optional): unit of the measurement result. Defaults to None.
            remark (str, optional): any kind of string remark about the measurement result. Defaults to None.
        """
        new_result = {
            "value": measurement_result,
            "parameter_name": parameter_name,
            "min": minimum,
            "max": maximum,
            "typ": typ,
            "unit": unit,
            "remark": remark
        }
        cls.results.append(new_result)
        if dump:
            result_write_object.append_result(
                test_ID='INFO',
                parameter_name=parameter_name,
                measured_value=measurement_result,
                minimum=minimum,
                maximum=maximum,
                typ=typ,
                unit=unit
            )


    @classmethod
    def AddPrimaryResult(
        cls, 
        test_ID: str,
        measurement_result,
        parameter_name: str,
        minimum,
        maximum,
        unit: str,
        typ=None,
        remark: str=None,
        dump: bool=False,
        result_write_object: result_file_writer=None
    ):
        """
        This method adds a new primery result to the test.
        All results are given as primary results have to be within range to pass the test

        Args:
            test_ID (str): ID of the test
            measurement_result (str|float|int): measurement result value
            parameter_name (str): name of ther parameter that the measurement result belongs to. Defaults to None.
            min (str|float|int): lower limit. Defaults to None.
            max (str|float|int): upper limit. Defaults to None.
            typ (str|float|int, optional): typical value. Defaults to None.
            unit (str): unit of the measurement result. Defaults to None.
            remark (str, optional): any kind of string remark about the measurement result. Defaults to None.
        """
        new_result = {
            "test_ID": test_ID,
            "value": measurement_result,
            "parameter_name": parameter_name,
            "min": minimum,
            "max": maximum,
            "typ": typ,
            "unit": unit,
            "remark": remark
        }
        cls.primary_results.append(new_result)
        if dump:
            result_write_object.append_result(
                test_ID='INFO',
                parameter_name=parameter_name,
                measured_value=measurement_result,
                minimum=minimum,
                maximum=maximum,
                typ=typ,
                unit=unit
            )


    @classmethod
    def EvaluateTest(cls, assertion=False) -> ResultEval:
        """
        This method checks all the primary results and sets the result of the test accordingly

        Args:
            assertion(bool): if it is set to true the evaluation raises assertion error

        Returns:
            ResultEval: judgement about the test result
        """
        final_result = True
        infotext = ''
        for result in cls.primary_results:
            eval_type = str(type(result["value"])).replace("<class '", '').replace("'>", '')
            eval_method = getattr(cls, f'_eval_{eval_type}')
            in_range = eval_method(
                cls,
                value=result["value"],
                minimum=result["min"],
                maximum=result["max"],
                unit=result["unit"]
            )
            act_info = f'{result["parameter_name"]}: {in_range} - {result["value"]} ( {result["min"]} / {result["max"]} )'
            my_logger.debug(act_info)
            infotext = infotext + '\n' + act_info
            final_result = final_result and in_range
        evaluation = cls.ResultEval.PASSED if final_result else cls.ResultEval.FAILED
        if assertion and evaluation == cls.ResultEval.FAILED:
            raise AssertionError(infotext)
        return evaluation

    def _eval_str(self, value: str, minimum: str, maximum: str, unit: str='hex') -> bool:
        """
        This method evaluates numbers flattened into string like hexadecimal string

        Args:
            value (str): string value
            min (str): string lower limit
            max (str): string upper limit
            unit (str): unit defines the format of the number formatted into string as:
                - hex : hexadecimal
                - bin : binary
                - dec : decimal

        Returns:
            bool: returns true if value is between the limits
        """
        my_conversion = getattr(self, f'_convert_{unit}')
        conv_value = my_conversion(value)
        conv_min = my_conversion(minimum)
        conv_max = my_conversion(maximum)
        return (conv_value >= conv_min and conv_value <= conv_max)


    @classmethod
    def _convert_hex(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 16)


    @classmethod
    def _convert_bin(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 2)


    @classmethod
    def _convert_dec(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 10)


    @classmethod
    def _remove_type_marker(self, string_in: str) -> str:
        """
        This method removes the possible type markers from the beginning of the numeric strings like 0x, 0b, 0d
        It also removes extra spaces from the string.

        Args:
            string_in (str): [description]

        Returns:
            str: [description]
        """
        my_string = string_in.replace('0x', '')
        my_string = my_string.replace('0b', '')
        my_string = my_string.replace('0d', '')
        my_string = my_string.replace('0u', '')
        my_string = my_string.replace(' ', '')
        return my_string


    def _eval_float(self, value: float, minimum: float, maximum: float, unit: str='V') -> bool:
        """
        This method evaluates a result that is floating point numeric

        Args:
            value (float): value
            min (float): lower limit
            max (float): upper limit
            unit (str, optional): unit of the physical quantity. Defaults to 'V'.

        Returns:
            bool: returns true if value is between the limits
        """
        return (value >= minimum and value <= maximum)


    def _eval_int(self, value: int, minimum: int, maximum: int, unit: str='V') -> bool:
        """
        This method evaluates a result that is integer

        Args:
            value (int): value
            min (int): lower limit
            max (int): upper limit
            unit (str, optional): unit of the physical quantity. Defaults to 'V'.

        Returns:
            bool: returns true if value is between the limits
        """
        return (value >= minimum and value <= maximum)


    @classmethod
    def AssertResults(cls):
        """
        This method carries out the evaluation of primary result by using assertpy assertion library
        """
        for result in cls.primary_results:
            eval_type = str(type(result["value"])).replace("<class '", '').replace("'>", '')
            eval_method = getattr(cls, f'_assert_{eval_type}')
            eval_method(
                cls,
                value=result["value"],
                minimum=result["min"],
                maximum=result["max"],
                unit=result["unit"]
            )


    def _assert_str(self, value: str, minimum: str, maximum: str, unit: str='hex'):
        """
        This method evaluates numbers flattened into string like hexadecimal string

        Args:
            value (str): string value
            min (str): string lower limit
            max (str): string upper limit
            unit (str): unit defines the format of the number formatted into string as:
                - hex : hexadecimal
                - bin : binary
                - dec : decimal
        """
        my_conversion = getattr(self, f'_convert_{unit}')
        conv_value = my_conversion(value)
        conv_min = my_conversion(minimum)
        conv_max = my_conversion(maximum)
        assert_that(conv_value).is_between(conv_min, conv_max)


    def _assert_float(self, value: float, minimum: float, maximum: float, unit: str='V'):
        """
        This method evaluates a result that is floating point numeric

        Args:
            value (float): value
            min (float): lower limit
            max (float): upper limit
            unit (str, optional): unit of the physical quantity. Defaults to 'V'.
        """
        assert_that(value).is_between(minimum, maximum)


    def _assert_int(self, value: float, minimum: float, maximum: float, unit: str='V'):
        """
        This method evaluates a result that is floating point numeric

        Args:
            value (float): value
            min (float): lower limit
            max (float): upper limit
            unit (str, optional): unit of the physical quantity. Defaults to 'V'.
        """
        my_logger.debug(value)
        my_logger.debug(minimum)
        my_logger.debug(maximum)
        assert_that(value).is_between(minimum, maximum)


class TestDetail:
    """
    This class specifies the type structure for test details that has to be given for all tests
    """
    test_ID: str=''
    Parameter_name: str=''
    Value: NoneType=None
    Evaluation: bool=True
    Unit: str=''
    Minimum: NoneType=None
    Maximum: NoneType=None
    Typ: NoneType=None


    @classmethod
    def SetDetails(
        cls,
        test_ID: str,
        Parameter_name: str,
        Value: NoneType=None,
        Evaluation: bool=True,
        Unit: str='',
        Minimum: NoneType=None,
        Maximum: NoneType=None,
        Typ: NoneType=None,

    ) -> None:
        """
        This method initializes the details for a new test

        Args:
            test_ID (str): the unique identifier of the test - HAS TO BE DEFINED
            Parameter_name (str): name of the parameter - HAS TO BE DEFINED
            Value (NoneType, optional): the outcome of the test - given in the Unit. Defaults to None.
            Evaluation (bool, optional): Decision whether the test is passed. Defaults to True.
            Unit (str, optional): unit the value is given in. Defaults to ''.
            Min (NoneType, optional): lower limit - given in the Unit. Defaults to None.
            Max (NoneType, optional): upper limit - given in the Unit. Defaults to None.
            Typ (NoneType, optional): typical value - given in the Unit. Defaults to None.
        """
        cls.test_ID = test_ID
        cls.Parameter_name = Parameter_name
        cls.Value = Value
        cls.Evaluation = Evaluation
        cls.Unit = Unit
        cls.Minimum = Minimum
        cls.Maximum = Maximum
        cls.Typ = Typ


    @classmethod
    def AddResult(cls, result: NoneType, dump: bool=True, result_writer_object: result_file_writer=None):
        """
        The outcome of the test can be given with this method

        Args:
            result (NoneType): the outcome of the test - given in the Unit
        """
        cls.Value = result
        if dump:
            result_writer_object.append_result(
                test_ID=cls.test_ID,
                parameter_name=cls.Parameter_name,
                measured_value=cls.Value,
                minimum=cls.Minimum,
                maximum=cls.Maximum,
                typ=cls.Typ,
                unit=cls.Unit
            )


    @classmethod
    def Evaluate(cls) -> bool:
        """
        This method evaluates the test and returns true if test is passed based on the given parameters

        Returns:
            bool: this is true if the test is passed
        """
        eval_type = str(type(cls.Value)).replace("<class '", '').replace("'>", '')
        eval_method = getattr(cls, f'_eval_{eval_type}')
        cls.Evaluation = eval_method(
            cls.Value,
            cls.Minimum,
            cls.Maximum,
            cls.Unit
        )
        if not cls.Evaluation:
            raise AssertionError(
                f'Test is failed, {cls.Value} {cls.Unit} is out of range {cls.Minimum} and {cls.Maximum}'
            )


    @classmethod
    def _eval_str(self, value: str, minimum: str, maximum: str, unit: str='hex') -> bool:
        """
        This method evaluates numbers flattened into string like hexadecimal string

        Args:
            value (str): string value
            min (str): string lower limit
            max (str): string upper limit
            unit (str): unit defines the format of the number formatted into string as:
                - hex : hexadecimal
                - bin : binary
                - dec : decimal

        Returns:
            bool: returns true if value is between the limits
        """
        my_conversion = getattr(self, f'_convert_{unit}')
        conv_value = my_conversion(value)
        conv_min = my_conversion(minimum)
        conv_max = my_conversion(maximum)
        return (conv_value >= conv_min and conv_value <= conv_max)


    @classmethod
    def _convert_hex(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 16)


    @classmethod
    def _convert_bin(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 2)


    @classmethod
    def _convert_dec(self, numeric_string: str) -> int:
        return int(self._remove_type_marker(numeric_string), 10)


    @classmethod
    def _remove_type_marker(self, string_in: str) -> str:
        """
        This method removes the possible type markers from the beginning of the numeric strings like 0x, 0b, 0d
        It also removes extra spaces from the string.

        Args:
            string_in (str): [description]

        Returns:
            str: [description]
        """
        my_string = string_in.replace('0x', '')
        my_string = my_string.replace('0b', '')
        my_string = my_string.replace('0d', '')
        my_string = my_string.replace('0u', '')
        my_string = my_string.replace(' ', '')
        return my_string


    @classmethod
    def _eval_float(self, value: float, minimum: float, maximum: float, unit: str='V') -> bool:
        """
        This method evaluates a result that is floating point numeric

        Args:
            value (float): value
            min (float): lower limit
            max (float): upper limit
            unit (str, optional): unit of the physical quantity. Defaults to 'V'.

        Returns:
            bool: returns true if value is between the limits
        """
        return (value >= minimum and value <= maximum)


    @classmethod
    def _eval_int(self, value: int, minimum: int, maximum: int, unit: str='V') -> bool:
        """
        This method evaluates a result that is integer

        Args:
            value (int): value
            min (int): lower limit
            max (int): upper limit
            unit (str, optional): unit of the physical quantity. Defaults to 'V'.

        Returns:
            bool: returns true if value is between the limits
        """
        return (value >= minimum and value <= maximum)

